/*Calculates size-adapted average bid-ask spread measures based on the functional forms provided for download.*/
*The user should adapt the path to the folder containing the data set with the downloaded functional form that we provide for download;

%let dir_functionalForm = F:\SizeAdaptedLiquidity\FunctionalForms;


PROC IMPORT OUT= SIZEADAP.AVGBIDASK_c_vol 
            DATAFILE= "&dir_functionalForm.\c_vol_avgbidask.csv" 
            DBMS=CSV REPLACE;
     GETNAMES=YES;
     DATAROW=2; 
RUN;
PROC IMPORT OUT= SIZEADAP.AVGBIDASK_norm_fact 
            DATAFILE= "&dir_functionalForm.\normalization_factor_avgbidask.csv" 
            DBMS=CSV REPLACE;
     GETNAMES=YES;
     DATAROW=2; 
RUN;


*create volume weighting scheme (see Internet Appendix C.1);
data vol_weight_seg;
input segment lower_bound upper_bound;
cards;
1 0 17500
2 17500 37500
3 37500 75000
4 75000 150000
5 150000 350000
6 350000 750000
7 750000 1500000
8 1500000 3500000
9 3500000 7500000
10 7500000 15000000
11 15000000 100000000000
run;

proc transpose data= SIZEADAP.AVGBIDASK_c_vol out=functional_form;
by entrd_vol_qt;
quit;

proc sql;
create table functional_form as 
select input(substr(_NAME_,7,4),4.) as year, input(substr(_NAME_,12,2),2.) as month, entrd_vol_qt, _NAME_, col1 as c_vol
from functional_form
order by _name_, entrd_vol_qt;
quit;

/*weight by inverse of the number of observations (see Appendix C.1)*/
proc sql;
create table groups as
select distinct  substr(s.CusipObservationDay,11,7) as ObservationMonth, g.*, count(*) as nobs, 1/count(*)*1000000 as weight_groups /*multiply with 1000000 for numerical reasons (the multiplier cancels out)*/
from vol_weight_seg g, SizeAdap.Avgbidask_sample s
where g.lower_bound<=s.entrd_vol_qt<g.upper_bound
group by calculated ObservationMonth, segment;
quit;

/*Calculate daily average bid-ask spread (left hand side of Equation (6))*/
proc means data=SizeAdap.Avgbidask_sample mean noprint;
	var rptd_pr;
	by CusipObservationDay rpt_side_cd;
	output out=AvgBidAsk_Means  mean(rptd_pr)=rptd_pr_mean;
run;

proc sql;
create table AvgBidAsk_Daily_BidAsk as
select a.CusipObservationDay, SUBSTR(a.CusipObservationDay, 1, 17) as CusipObservationMonth,
	2*max(a.rptd_pr_mean - b.rptd_pr_mean,0)/(a.rptd_pr_mean + b.rptd_pr_mean) as rel_bidAsk_daily
from AvgBidAsk_Means a, AvgBidAsk_Means b
where a.CusipObservationDay=b.CusipObservationDay
and a.rpt_side_cd='S' and b.rpt_side_cd='B'
order by SUBSTR(a.CusipObservationDay, 1, 17);
quit;

proc sql;
create table avgBAS_2nd_step_input as
select distinct s.CusipObservationDay,r.CusipObservationMonth, s.rpt_side_cd,  min(s.entrd_vol_qt,20000000) as entrd_vol_qt, 
	count(*)/m._FREQ_ as weight, r.rel_bidAsk_daily
from SizeAdap.Avgbidask_sample s, AvgBidAsk_Means m, AvgBidAsk_Daily_BidAsk r
where s.CusipObservationDay=m.CusipObservationDay and s.rpt_side_cd=m.rpt_side_cd
	and s.CusipObservationDay=r.CusipObservationDay
group by s.CusipObservationDay, s.rpt_side_cd, calculated entrd_vol_qt;
quit;


proc sql;
create table avgBAS_2nd_step_input as
select distinct a.cusipObservationDay, a.cusipObservationMonth, a.rel_bidAsk_daily, a.entrd_vol_qt, a.rpt_side_cd, 
	a.weight, b.segment, b.weight_groups, sum(a.weight*b.weight_groups)/sum(a.weight) as cusipObsDay_weight
from avgBAS_2nd_step_input a, groups b
where b.lower_bound<=a.entrd_vol_qt<b.upper_bound 
	and b.ObservationMonth=substr(a.CusipObservationDay,11,7)
group by a.cusipObservationDay;
quit;

proc sql;
create table avgBAS_2nd_step_input as
select distinct x.CusipObservationDay, x.CusipObservationMonth, x.rel_bidAsk_daily, 
	sum(1/2*x.weight*y.c_vol) as tc_estimate_avg, x.cusipObsDay_weight
from avgBAS_2nd_step_input x, functional_form y
where x.entrd_vol_qt=y.entrd_vol_qt
	and y.year = input(substr(x.cusipObservationMonth,11,4),4.) and y.month = input(substr(x.cusipObservationMonth,16,2),2.)
group by x.cusipObservationDay;
quit;

*calculate size-adapted measure;
options nonotes nosource nosource2;
proc nlin data=avgBAS_2nd_step_input Outest=reg_2nd_step noprint; 
parms 
	c = 0;
	model rel_bidAsk_daily = c*tc_estimate_avg;
_weight_ =cusipObsDay_weight;
bounds c >= 0;
by CusipObservationMonth;
run;
options notes source source2;

*Save final results; 
proc sql;
create table sizeAdap.avgBidAsk_A_size_results as
select r.CusipObservationMonth, r.c * n.normalization_factor as SizeAdap_AvgBidAsk
from reg_2nd_step as r, SizeAdap.Avgbidask_norm_fact n
where r._TYPE_ = 'FINAL'
	and input(substr(r.cusipObservationMonth,11,4),4.)=n.year and  input(substr(r.cusipObservationMonth,16,2),2.)=n.month
order by CusipObservationMonth;
quit; 

