/*Calculates size-adapted average bid-ask spread measures based on the iterative procedure described in the paper.*/

*create volume weighting scheme (see Internet Appendix C.1);
data vol_weight_seg;
input segment lower_bound upper_bound;
cards;
1 0 17500
2 17500 37500
3 37500 75000
4 75000 150000
5 150000 350000
6 350000 750000
7 750000 1500000
8 1500000 3500000
9 3500000 7500000
10 7500000 15000000
11 15000000 100000000000
run;

*get all traded volumes. The functional form is evaluated at these points. It would be possible to include additional volumes here.;
proc sql;
create table traded_vols as
select distinct min(entrd_vol_qt,20000000) as entrd_vol_qt
from SizeAdap.AvgBidAsk_Sample;
quit;

/*weight by inverse of the number of observations (see Appendix C.1)*/
proc sql;
create table groups as
select distinct  substr(s.CusipObservationDay,11,7) as ObservationMonth, g.*, count(*) as nobs, 1/count(*)*1000000 as weight_groups /*multiply with 1000000 for numerical reasons (the multiplier cancels out)*/
from vol_weight_seg g, SizeAdap.Avgbidask_sample s
where g.lower_bound<=s.entrd_vol_qt<g.upper_bound
group by calculated ObservationMonth, segment;
quit;

/*Calculate daily average bid-ask spread (left hand side of Equation (6))*/
proc means data=SizeAdap.Avgbidask_sample mean noprint;
	var rptd_pr;
	by CusipObservationDay rpt_side_cd;
	output out=AvgBidAsk_Means  mean(rptd_pr)=rptd_pr_mean;
run;

proc sql;
create table AvgBidAsk_Daily_BidAsk as
select a.CusipObservationDay, SUBSTR(a.CusipObservationDay, 1, 17) as CusipObservationMonth,
	2*max(a.rptd_pr_mean - b.rptd_pr_mean,0)/(a.rptd_pr_mean + b.rptd_pr_mean) as rel_bidAsk_daily
from AvgBidAsk_Means a, AvgBidAsk_Means b
where a.CusipObservationDay=b.CusipObservationDay
and a.rpt_side_cd='S' and b.rpt_side_cd='B'
order by SUBSTR(a.CusipObservationDay, 1, 17);
quit;


proc sql;
create table avgBidAsk_freq as
select distinct s.CusipObservationDay,r.CusipObservationMonth, s.rpt_side_cd, s.entrd_vol_qt, 
	count(*) as freq_vol, m._FREQ_ as freq_total label='', count(*)/m._FREQ_ as weight,
	r.rel_bidAsk_daily
from SizeAdap.AvgBidAsk_Sample s, AvgBidAsk_Means m, Avgbidask_daily_bidask r
where s.CusipObservationDay=m.CusipObservationDay and s.rpt_side_cd=m.rpt_side_cd
	and s.CusipObservationDay=r.CusipObservationDay
group by s.CusipObservationDay, s.rpt_side_cd, s.entrd_vol_qt;
quit;

*Calculate average daily bid and ask prices and average daily bid-ask spreads for each combination of cusipObservationDay and volume 
 (only applicable for cusipObservationDays on which different volumes are traded) 
 (see Internet Appendix C.2);
proc sql;
create table AvgBidAsk_Sample_sub as
select cusipObservationDay, rptd_pr, rpt_side_cd, trd_exctn_dt, min(entrd_vol_qt,20000000) as entrd_vol_qt, log_vol, dollar_vol, log_dollar_vol
from SizeAdap.AvgBidAsk_Sample
group by cusipObservationDay
having count(distinct min(entrd_vol_qt,20000000))>1;
quit;
proc sql;
create table AvgBidAsk_Sample_sub as
select *
from AvgBidAsk_Sample_sub
group by cusipObservationDay, entrd_vol_qt
having count(distinct rpt_side_cd)>1
order by cusipObservationDay, entrd_vol_qt, rpt_side_cd;
quit;
proc means data=AvgBidAsk_Sample_sub mean noprint;
	var rptd_pr;
	by CusipObservationDay entrd_vol_qt rpt_side_cd;
	output out=AvgBidAsk_Means  mean(rptd_pr)=rptd_pr_mean;
run;
proc sql;
create table AvgBidAsk_Daily_BidAsk as
select a.CusipObservationDay, SUBSTR(a.CusipObservationDay, 1, 17) as CusipObservationMonth, a.entrd_vol_qt,
	2*max(a.rptd_pr_mean - b.rptd_pr_mean,0)/(a.rptd_pr_mean + b.rptd_pr_mean) as rel_bidAsk_daily
from AvgBidAsk_Means a, AvgBidAsk_Means b
where a.CusipObservationDay=b.CusipObservationDay and a.entrd_vol_qt=b.entrd_vol_qt
	and a.rpt_side_cd='S' and b.rpt_side_cd='B'
order by a.CusipObservationDay;
quit;
proc sql;
create table avgBidAsk_freq_sub as
select distinct s.CusipObservationDay, r.CusipObservationMonth, s.rpt_side_cd, s.entrd_vol_qt, 
	count(*) as freq_vol, r.rel_bidAsk_daily
from AvgBidAsk_Sample_sub s, AvgBidAsk_Means m, Avgbidask_daily_bidask r
where s.CusipObservationDay=m.CusipObservationDay and s.entrd_vol_qt=m.entrd_vol_qt and s.rpt_side_cd=m.rpt_side_cd
	and s.CusipObservationDay=r.CusipObservationDay and s.entrd_vol_qt=r.entrd_vol_qt
group by s.CusipObservationDay, s.rpt_side_cd, s.entrd_vol_qt;
quit;
proc sql;
create table avgBidAsk_freq_sub as
select CusipObservationDay, CusipObservationMonth, rpt_side_cd, entrd_vol_qt, freq_vol, sum(freq_vol) as freq_total, freq_vol/(sum(freq_vol)) as weight, 
	/*weight is the ratio of how often this volume is traded compared to all traded volumes on the respective day for which we can calculate the 
	size-dependent bid-ask spread (see C.2)*/
	rel_bidAsk_daily
from avgBidAsk_freq_sub
group by cusipObservationDay, rpt_side_cd;
quit;

*Create auxiliary tables to put everything together for the input used in the iterative procedure
 (aux_1 is based on the daily bid-ask spreads, aux_2 on the bid-ask spreads per day and volume);
proc sql;
create table aux_1 as
select a.cusipObservationDay,
	case when max(a.entrd_vol_qt)=min(a.entrd_vol_qt) then 1 else 1*10**(-12) end as bandw_mod, 
		/*mark trades for which not all volumes are the same with a small number (done for reasons of programming efficiency), 
		the exact number (i.e., 1*10**(-12) has no impact on the results)*/
	sum(a.weight*b.weight_groups)/sum(a.weight) as cusipObsDay_weight
from avgBidAsk_freq a, groups b
where b.lower_bound<=a.entrd_vol_qt<b.upper_bound 
	and b.ObservationMonth=substr(a.CusipObservationDay,11,7)
group by a.cusipObservationDay;
quit;
proc sql;
create table aux_2 as
select distinct cusipObservationDay, entrd_vol_qt, rel_bidAsk_daily, 1/2*sum(weight) as vol_weight 
	/*calculate the final ratio of how often a volume trades relative to the other volumes (formula in the second paragraph of Appendix C.2)*/
from avgBidAsk_freq_sub
group by cusipObservationDay, entrd_vol_qt;
quit;

*Create regression input set for monthly iterative procedure;
proc sql;
create table sizeAdap.avgBidAsk_regInput_mth as
select cusipObservationDay, cusipObservationMonth, rel_bidAsk_daily, min(entrd_vol_qt,20000000) as entrd_vol_qt, log(min(entrd_vol_qt,20000000)) as log_vol, weight
from avgBidAsk_freq;
quit;
proc sql;
create table sizeAdap.avgBidAsk_regInput_mth as
select distinct a.cusipObservationDay, a.cusipObservationMonth, a.rel_bidAsk_daily, a.entrd_vol_qt, a.log_vol, b.segment,
	1/2*sum(a.weight) as vol_weight, b.weight_groups
	/*sum up the weight per volume for both sides*/
from sizeAdap.avgBidAsk_regInput_mth a, groups b
where b.lower_bound<=a.entrd_vol_qt<b.upper_bound 
	and b.ObservationMonth=substr(a.CusipObservationDay,11,7)
group by a.cusipObservationDay, a.entrd_vol_qt;
quit;
/*add information from auxiliary tables*/
proc sql;
create table sizeAdap.avgBidAsk_regInput_mth as
select a.*, h.bandw_mod, h.cusipObsDay_weight
from sizeAdap.avgBidAsk_regInput_mth a, aux_1 h
where a.cusipObservationDay=h.cusipObservationDay;
quit;
proc sql;
create table sizeAdap.avgBidAsk_regInput_mth as
select a.*, h.rel_bidAsk_daily as rel_bidAsk_vol, h.vol_weight as vol_weight_new
from sizeAdap.avgBidAsk_regInput_mth a
left join aux_2 h
on a.cusipObservationDay=h.cusipObservationDay and a.entrd_vol_qt=h.entrd_vol_qt;
quit;
/*combine information from the two cases
	1) days on which multiple volumes are traded (vol_weight_new not NULL) 
	2) day on which multiple volumes are not traded (vol_weight_new is NULL as no entry is in aux_2). 
 The final weight is calculated as the product of the volume category weight and the ratio of how often this volume 
is traded compared to all traded volumes on the respective day for which we can calculate the size-dependent bid-ask 
spread (see C.2)*/
proc sql;
create table sizeAdap.avgBidAsk_regInput_mth as
select *, case when vol_weight_new ne . then rel_bidAsk_vol else rel_bidAsk_daily end as rel_bidAsk_1stage,
	case when vol_weight_new ne . then vol_weight_new*weight_groups else bandw_mod*weight_groups end as obs_weight_1stage
from sizeAdap.avgBidAsk_regInput_mth;
quit;

*Add month and year to data set;
proc sql;
create table sizeAdap.avgBidAsk_regInput_mth as
select input(substr(cusipObservationMonth,11,4),4.) as year,  input(substr(cusipObservationMonth,16,2),2.) as month, *
from sizeAdap.avgBidAsk_regInput_mth;
quit;

*create functional form data set;
proc sql;
create table sizeAdap.avgBidAsk_c_vol_iter as
select *
from traded_vols;
quit;


*delete old factor file (in case the algorithm has been executed before);
proc datasets library=sizeAdap;
delete avgBidAsk_monthly;
quit;


*define iteration macro;
%macro start_iteration();

	*variable to keep track on the number of observations in the current and in the last run;
	%local avg_abs_change;
	%let avg_abs_change=100.5; /*starting value to keep track of the changes (use a large number)*/
	%local conversion_goal;
	%let conversion_goal=0.000001;
	%put &avg_abs_change;
	%do %while (%SYSEVALF(&avg_abs_change-&conversion_goal,floor)>=0);
	%put &avg_abs_change;
		
		*Create 1st-stage regression input;
		/*as on the right side of the non-parametric regression, we can just have the functional form,
		we divide the left-hand side (i.e., the bid-ask spread) by the factor and adjust the weighting
		accordingly (i.e., dividing the regression equation by a factor X leads to the squared residual 
		divided by X^2. To adjust for this, we have to multiply the weight of the observation by X^2).*/
		proc sql;
		create table regression_1st_input as
		select x.cusipObservationDay, x.cusipObservationMonth, x.rel_bidAsk_1stage, x.entrd_vol_qt, x.segment, y.factor, 
			x.vol_weight, x.vol_weight_new, x.rel_bidAsk_1stage/y.factor as rel_bidAsk_adj, 
			x.obs_weight_1stage, x.obs_weight_1stage*(y.factor**2) as obs_weight_Adj
		from regression_1st_input x, Final_set y
		where x.CusipObservationMonth=y.CusipObservationMonth and abs(y.factor)>10**(-7)
		order by cusipobservationday;
		quit;

		proc sql;
		create table avgBidAsk_1stage_missing_vols as
		select *
		from traded_vols
		where entrd_vol_qt not in (select entrd_vol_qt from regression_1st_input);
		quit;

		*first-stage regression;
		proc loess data=regression_1st_input;
			model rel_bidAsk_adj=entrd_vol_qt / smooth=0.3;
			score data=avgBidAsk_1stage_missing_vols;
			weight obs_weight_Adj;
			ods output OutputStatistics=regression_1st;
			ods Output ScoreResults=regression_1st_score;
		run;

		proc sql;
		create table regression_1st_res as
		select distinct entrd_vol_qt, pred as c_vol
		from regression_1st
		union
		select entrd_vol_qt, p_rel_bidAsk_adj as c_vol
		from regression_1st_score;
		quit;

		*Create input for second-stage regression and run 2nd-stage regression;
		proc sql;
		create table regression_2nd_input as
		select distinct x.year, x.month, x.CusipObservationDay, x.CusipObservationMonth, x.rel_bidAsk_daily, 
			sum(x.vol_weight*y.c_vol) as tc_estimate_avg, x.cusipObsDay_weight
		from regInput x, regression_1st_res y
		where x.entrd_vol_qt=y.entrd_vol_qt
		group by x.cusipObservationDay;
		quit;
		
		options nonotes nosource nosource2;
		proc nlin data=regression_2nd_input Outest=regression noprint; 
		parms 
			c = 0;
			model rel_bidAsk_daily = c*tc_estimate_avg;
		_weight_ =cusipObsDay_weight;
		bounds c >= 0;
		by CusipObservationMonth;
		run;
		options notes source source2;
		
		*scale results so that the average factor is 1 (see second paragraph of Internet Appendix C.1).
		 "factor" is the "normalized scaling factor" in the paper.;
		proc sql;
		create table aux_2nd_reg as
		select r.CusipObservationMonth, r.c as factor
		from regression as r
		where r._TYPE_ = 'FINAL';
		quit; 
		
		proc sql noprint;
		select distinct avg(factor)
		into :avg_fac
		from aux_2nd_reg;
		quit;

		proc sql;
		create table aux_2nd_reg as
		select CusipObservationMonth, factor/&avg_fac as factor
		from aux_2nd_reg;
		quit;

		proc sql;
		create table regression_1st_res as
		select entrd_vol_qt, c_vol*&avg_fac as c_vol
		from regression_1st_res;
		quit;

		*Get differences to previous iteration, also save SSE difference;
		proc sql;
		create table abs_diff as
		select r.CusipObservationMonth, r.factor, o.factor as factor_old, abs(o.factor-r.factor) as abs_diff
		from aux_2nd_reg as r, Final_set o
		where r.CusipObservationMonth=o.CusipObservationMonth;
		quit;
	
		proc sql;
		create table SSE_diff as
		select distinct o.sum_sse-sum(y._SSE_) as SSE_diff 
		from sum_squares o, regression y 
		where y._type_ = 'FINAL';
		quit;
	
		proc sql;
		create table SSE_diff as
		select distinct a.*, avg(b.abs_diff) as avg_fac_diff
		from SSE_diff a, abs_diff b;
		quit;

		proc sql;
		create table sum_squares as
		select sum(y._SSE_) as sum_sse, count(*) as number_of_observations
		from regression y
		where y._type_ = 'FINAL';
		quit;

		*Save iteration results data set and summary data;
		proc sql;
		create table Final_set as
		select r.CusipObservationMonth, r.factor
		from aux_2nd_reg r;
		quit;

		data convergence_avgBidAsk_&cur_year._&cur_mth.;
			set convergence_avgBidAsk_&cur_year._&cur_mth. sum_squares SSE_diff;
		run;

		*Update avg_abs_change;
		PROC SQL; 
	  		SELECT DISTINCT avg_fac_diff 
	   		INTO :avg_abs_change
	    FROM SSE_diff; 
	 	QUIT;

	%put &avg_abs_change;

%end;
%mend start_iteration;

%macro monthly(cur_year,cur_mth);

/*initiate iteration macro for each month*/
/*select the end of the sample period (integer number for the number of months): 
year * 12 + month + 1, i.e., for June 2020: 2020 * 12 + 6 + 1 = 24247*/
%do %while (%eval(&cur_year*12+&cur_mth)<24247);
	%put &cur_year., &cur_mth.;

	/*set input data set to current observation month*/
	proc sql;
	create table regInput as
	select *
	from sizeAdap.Avgbidask_reginput_mth
	where year=&cur_year and month=&cur_mth;
	quit;

	/*create regression input data for 1st stage and exclude missing volumes, for which bandw_mod was set to 1*10**(-12). 
	For those, the weight multiplied by 1*10**(-12)), is smaller than 10**(-7))*/
	proc sql;
	create table regression_1st_input as
	select *
	from regInput
	where obs_weight_1stage>10**(-7); 
	quit;

	/*get the volumes that are not in the dataset for the current month (but in the overall dataset)*/
	proc sql;
	create table avgBidAsk_1stage_missing_vols as
	select *
	from traded_vols
	where entrd_vol_qt not in (select entrd_vol_qt from regression_1st_input);
	quit;

	/*Clear log so that it does not overflow*/
	DM 'Clear log' ; 

	/*1st regression (starting value for factor=1). 
	  Smoothing parameter is set to 30% (see discussion in first paragraph of Internet Appendix C.1)*/
	proc loess data=regression_1st_input;
		model rel_bidAsk_1stage=entrd_vol_qt / smooth=0.3;
		score data=avgBidAsk_1stage_missing_vols;
		weight obs_weight_1stage;
		ods output OutputStatistics=regression_1st;
		ods Output ScoreResults=regression_1st_score;
	run;

	proc sql;
	create table regression_1st_res as
	select distinct entrd_vol_qt, pred as c_vol
	from regression_1st
	union
	select entrd_vol_qt, p_rel_bidAsk_1stage as c_vol
	from regression_1st_score;
	quit;

	/*Create input and output of 2nd regression*/
	proc sql;
	create table regression_2nd_input as
	select distinct x.CusipObservationDay, x.CusipObservationMonth, x.rel_bidAsk_daily, 
		sum(x.vol_weight*y.c_vol) as tc_estimate_avg, x.cusipObsDay_weight
	from regInput x, regression_1st_res y
	where x.entrd_vol_qt=y.entrd_vol_qt
	group by x.cusipObservationDay;
	quit;

	/*2nd step regression*/
	options nonotes nosource nosource2;
	proc nlin data=regression_2nd_input Outest=regression noprint; 
	/* rel_Delta for relative measure */
	parms 
		c = 0;
		model rel_bidAsk_daily = c*tc_estimate_avg;
	_weight_ =cusipObsDay_weight;
	/* no negative spreads allowed */
	bounds c >= 0;
	by CusipObservationMonth;
	run;
	options notes source source2;
	
	/*save results of iteration */
	proc sql;
	create table Final_set as
	select r.CusipObservationMonth, r.c as factor
	from regression as r
	where r._TYPE_ = 'FINAL';
	quit;

	/*get average factor and scale final data sets such that the average factor equals 1 (see C.1)*/
	%global avg_fac;
	proc sql noprint;
	select distinct avg(factor)
	into :avg_fac
	from Final_set;
	quit;
	%put &avg_fac;

	proc sql;
	create table Final_set as
	select CusipObservationMonth, factor/&avg_fac as factor
	from Final_set;
	quit;

	proc sql;
	create table regression_1st_res as
	select entrd_vol_qt, c_vol*&avg_fac as c_vol
	from regression_1st_res;
	quit;

	proc sql;
		create table sum_squares as
		select sum(_SSE_)  as sum_sse, count(*) as number_of_observations
		from regression y
		where y._type_ = 'FINAL';
	quit;

	data convergence_avgBidAsk_&cur_year._&cur_mth.;
		set sum_squares;
	run;

	/*call the iteration macro*/
	%start_iteration();

	/*add functional from to data set*/
	proc sql;
	create table sizeAdap.avgBidAsk_c_vol_iter as
	select a.*, b.c_vol as c_vol_&cur_year._&cur_mth.
	from sizeAdap.avgBidAsk_c_vol_iter a, regression_1st_res b
	where a.entrd_vol_qt=b.entrd_vol_qt;
	quit;

	/*add factors to data set*/
	proc append base=sizeAdap.avgBidAsk_monthly data=Final_set;quit;


	/*update year and mth*/
	%if &cur_mth=12%then %do;
		%let cur_year=%eval(&cur_year+1);
		%let cur_mth=1;
	%end;
	%else %do;
		%let cur_mth=%eval(&cur_mth+1);
	%end;

%end;

%mend monthly;

*call macro;
%monthly(2004,10);

%global note;
%let note='Iteration finished';
%put &note;

/*Calculate normalization factor*/
proc sql;
create table vol_weight as
select distinct year(trd_exctn_dt) as year, month(trd_exctn_dt) as month, min(entrd_vol_qt,20000000) as entrd_vol_qt,
	sum(min(entrd_vol_qt,20000000)) as sum_vol
from sizeAdap.AvgBidAsk_Sample
group by calculated year, calculated month, calculated entrd_vol_qt;
quit;
proc sql;
create table vol_weight_2 as
select *, sum_vol/sum(sum_vol) as weight_vol
from vol_weight
group by year, month;
quit;

proc transpose data=sizeAdap.avgBidAsk_c_vol_iter out=c_vol_long (rename=(col1=c_vol)); by entrd_vol_qt;quit;

proc sql;
create table sizeAdap.Avgbidask_norm_fact_iter as
select distinct v.year, v.month, sum(v.weight_vol*f.c_vol) as normalization_factor
from vol_weight_2 v, c_vol_long f
where v.year=input(substr(f._NAME_,7,4),4.) and v.month=input(substr(f._NAME_,12,2),2.) and v.entrd_vol_qt=f.entrd_vol_qt
group by v.year, v.month;
quit;

*Save final results; 
proc sql;
create table sizeAdap.AvgbidAsk_b_size_results as
select r.CusipObservationMonth, r.factor * n.normalization_factor as SizeAdap_AvgBidAsk
from sizeAdap.avgBidAsk_monthly as r, SizeAdap.Avgbidask_norm_fact_iter n
where input(substr(r.cusipObservationMonth,11,4),4.)=n.year and  input(substr(r.cusipObservationMonth,16,2),2.)=n.month
order by CusipObservationMonth;
quit; 
