/*Calculates size-adapted Schultz measures based on the iterative procedure described in the paper.*/

*create volume weighting scheme (see Internet Appendix C.1);
data vol_weight_seg;
input segment lower_bound upper_bound;
cards;
1 0 17500
2 17500 37500
3 37500 75000
4 75000 150000
5 150000 350000
6 350000 750000
7 750000 1500000
8 1500000 3500000
9 3500000 7500000
10 7500000 15000000
11 15000000 100000000000
run;

*get all traded volumes. The functional form is evaluated at these points. It would be possible to include additional volumes here.;
proc sql;
create table traded_vols as
select distinct min(entrd_vol_qt,20000000) as entrd_vol_qt
from SizeAdap.Schultz_sample;
quit;

/*weight by inverse of the number of observations (see Appendix C.1)*/
proc sql;
create table groups as
select distinct substr(s.CusipObservationDay,11,7) as ObservationMonth, g.*, count(*) as nobs, 1/count(*) * 1000000 as weight_groups /*multiply with 1000000 for numerical reasons (the multiplier cancels out)*/
from vol_weight_seg g, SizeAdap.Schultz_sample s
where g.lower_bound<=s.entrd_vol_qt<g.upper_bound
group by calculated ObservationMonth, segment;
quit;

*Create regression input set for monthly iterative procedure;
proc sql;
create table sizeAdap.schultz_reginput_mth as
select s.cusipObservationDay, s.cusipObservationMonth, min(s.entrd_vol_qt,20000000) as entrd_vol_qt, s.rel_delta, s.Dummy, g.weight_groups, g.segment
from SizeAdap.Schultz_sample s, groups g
where g.lower_bound<=s.entrd_vol_qt<g.upper_bound
	and g.ObservationMonth=substr(s.CusipObservationDay,11,7)
order by cusipObservationDay, entrd_vol_qt, rel_delta, Dummy;
quit;

*Add month and year to data set;
proc sql;
create table sizeAdap.schultz_reginput_mth as
select input(substr(cusipObservationMonth,11,4),4.) as year, input(substr(cusipObservationMonth,16,2),2.) as month, *
from sizeAdap.schultz_reginput_mth;
quit;

*create functional form data set;
proc sql;
create table sizeAdap.Schultz_c_vol_iter as
select *
from traded_vols;
quit;

*delete old factor file (in case the algorithm has been executed before);
proc datasets library=sizeAdap;
delete Schultz_monthly;
quit;

*define iteration macro;
%macro start_iteration();

	*variable to keep track on the number of observations in the current and in the last run;
	%local iter;
	%let iter=0;
	%local avg_abs_change;
	%let avg_abs_change=100.5;/*starting value to keep track of the changes (use a large number)*/
	%local conversion_goal;
	%let conversion_goal=0.000001;
	%put &avg_abs_change;
	%do %while (%SYSEVALF(&avg_abs_change-&conversion_goal,floor)>=0 and &iter<=500);
	%put &avg_abs_change;


		*Create 1st-stage regression input;
		/*as on the right side of the non-parametric regression, we can just have the functional form,
		we divide the left-hand side (i.e., the bid-ask spread) by the factor and adjust the weighting
		accordingly (i.e., dividing the regression equation by a factor X leads to the squared residual 
		divided by X^2. To adjust for this, we have to multiply the weight of the observation by X^2).*/		
		proc sql;
		create table schultz_reg_1st_input as
		select x.entrd_vol_qt, x.rel_delta, x.Dummy, y.c_0, y.factor, (x.rel_delta-y.c_0)/(y.factor*x.Dummy) as reg_left_side, 
			x.weight_groups*(y.factor**2) as obs_weight
		from schultz_regInput x, Final_set y
		where x.CusipObservationMonth=y.CusipObservationMonth
			and x.dummy^=0 and abs(y.factor)>10**(-7)
		order by x.cusipObservationDay, x.entrd_vol_qt, x.rel_delta, x.Dummy;
		quit;
			
		proc sql;
		create table schultz_reg1st_missing_vols as
		select *
		from traded_vols
		where entrd_vol_qt not in (select entrd_vol_qt from schultz_reg_1st_input);
		quit;
		
		*first-stage regression;
		proc loess data=schultz_reg_1st_input;
			model reg_left_side=entrd_vol_qt / smooth=0.3;
			weight obs_weight;
			score data=schultz_reg1st_missing_vols;
			ods output OutputStatistics=regression_1st;
			ods Output ScoreResults=regression_1st_score;
		run;

		proc sql;
		create table regression_1st_res as
		select distinct entrd_vol_qt, pred as c_vol
		from regression_1st
		union
		select entrd_vol_qt, p_reg_left_side as c_vol
		from regression_1st_score;
		quit;

		*Create input for second-stage regression and run 2nd-stage regression;
		proc sql;
		create table schultz_reg_2nd_input as
		select s.cusipObservationMonth, s.entrd_vol_qt, s.Dummy, s.cusipObservationDay, r.c_vol as tc_estimate label='',
			s.rel_delta, s.weight_groups
		from schultz_regInput s, regression_1st_res r
		where s.entrd_vol_qt=r.entrd_vol_qt
		order by s.cusipObservationDay, s.entrd_vol_qt, s.rel_delta, s.Dummy;
		quit;		
		
		options nonotes nosource nosource2;
		proc nlin data=schultz_reg_2nd_input method=marquardt Outest=regression noprint; 
		parms c = 0	c_0=0;
			model rel_Delta = c_0 +  c*(Dummy*tc_estimate);
		_weight_ =weight_groups;
		bounds c >= 0;
		by CusipObservationMonth;
		run;
		options notes source source2;

		*scale results so that the average factor is 1 (see second paragraph of Internet Appendix C.1).
		 "factor" is the "normalized scaling factor" in the paper.;
		proc sql;
		create table aux_2nd_reg as
		select r.CusipObservationMonth, r.c_0, r.c as factor
		from regression as r
		where r._TYPE_ = 'FINAL';
		quit; 
		
		proc sql noprint;
		select distinct avg(factor)
		into :avg_fac
		from aux_2nd_reg;
		quit;

		proc sql;
		create table aux_2nd_reg as
		select CusipObservationMonth, c_0, factor/&avg_fac as factor
		from aux_2nd_reg;
		quit;

		proc sql;
		create table regression_1st_res as
		select entrd_vol_qt, c_vol*&avg_fac as c_vol
		from regression_1st_res;
		quit;

		*Get differences to previous iteration, also save SSE difference;
		proc sql;
		create table abs_diff as
		select r.CusipObservationMonth, r.c_0, r.factor, o.factor as factor_old, abs(o.factor-r.factor) as abs_diff
		from aux_2nd_reg as r, Final_set o
		where r.CusipObservationMonth=o.CusipObservationMonth;
		quit;

		proc sql;
		create table SSE_diff as
		select distinct o.sum_sse-sum(y._SSE_) as SSE_diff
		from sum_squares o, regression y
		where y._type_ = 'FINAL';
		quit; 
		
		proc sql;
		create table SSE_diff as
		select distinct a.*, max(b.abs_diff) as max_fac_diff, avg(b.abs_diff) as avg_fac_diff 
		from SSE_diff a, abs_diff b;
		quit;

		proc sql;
		create table sum_squares as
		select sum(y._SSE_) as sum_sse, count(*) as number_of_observations
		from regression y
		where y._type_ = 'FINAL';
		quit;

		*save iteration results data set and summary data;
		proc sql;
		create table final_set as
		select r.CusipObservationMonth, r.c_0, r.factor
		from aux_2nd_reg r;
		quit;

		data convergence_Schultz_&cur_year._&cur_mth.;
			set convergence_Schultz_&cur_year._&cur_mth. sum_squares SSE_diff;
		run;

		PROC SQL; 
	  		SELECT DISTINCT avg_fac_diff 
	   		INTO :avg_abs_change
	    FROM SSE_diff; 
	 	QUIT;
		
		%let iter=%eval(&iter+1);

	%put &avg_abs_change;

%end;
%mend start_iteration;

%macro monthly(cur_year,cur_mth);

/*initiate iteration macro for each month*/
/*select the end of the sample period (integer number for the number of months): 
year * 12 + month + 1, i.e., for June 2020: 2020 * 12 + 6 + 1 = 24247*/
%do %while (%eval(&cur_year*12+&cur_mth)<24247);
	%put &cur_year., &cur_mth.;

	/*set input data set to current observation month*/
	proc sql;
	create table schultz_regInput as
	select *
	from sizeAdap.schultz_regInput_mth
	where year=&cur_year and month=&cur_mth;
	quit;

	/*create regression input data for 1st stage*/
	proc sql;
	create table schultz_reg_1st_input as
	select entrd_vol_qt, rel_delta, Dummy, (rel_delta-0)/(1*Dummy) as reg_left_side, weight_groups
	from schultz_regInput
	where dummy^=0;
	quit;

	/*get the volumes that are not in the dataset for the current month (but in the overall dataset)*/
	proc sql;
	create table schultz_reg1st_missing_vols as
	select *
	from traded_vols
	where entrd_vol_qt not in (select entrd_vol_qt from schultz_reg_1st_input);
	quit;

	/*Clear log so that it does not overflow*/
	DM 'Clear log' ; 

	/*1st regression (starting value for factor=1). 
	Smoothing parameter is set to 30% (see discussion in first paragraph of Internet Appendix C.1)*/
	proc loess data=schultz_reg_1st_input;
		model reg_left_side=entrd_vol_qt / smooth=0.3;
		weight weight_groups;
		score data=schultz_reg1st_missing_vols;
		ods output OutputStatistics=regression_1st;
		ods Output ScoreResults=regression_1st_score;
	run;

	proc sql;
	create table regression_1st_res as
	select distinct entrd_vol_qt, pred as c_vol
	from regression_1st
	union
	select entrd_vol_qt, p_reg_left_side as c_vol
	from regression_1st_score;
	quit;

	/*Create input and output of 2nd regression*/
	proc sql;
	create table schultz_reg_2nd_input as
	select s.cusipObservationMonth, s.entrd_vol_qt, s.Dummy, s.cusipObservationDay, r.c_vol as tc_estimate label='', s.rel_delta, s.weight_groups
	from schultz_regInput s, regression_1st_res r
	where s.entrd_vol_qt=r.entrd_vol_qt
	order by s.cusipObservationDay, s.entrd_vol_qt, s.rel_delta, s.Dummy;
	quit;

	/*2nd step regression*/
	options nonotes nosource nosource2;
	proc nlin data=schultz_reg_2nd_input method=marquardt Outest=regression noprint; 
	/* rel_Delta for relative measure */
	parms c = 0	c_0=0;
		model rel_Delta = c_0 +  c*(Dummy*tc_estimate);
	_weight_ =weight_groups;
	/* no negative spreads allowed */
	bounds c >= 0;
	by CusipObservationMonth;
	run;
	options notes source source2;

	/*save results of iteration */
	proc sql;
	create table Final_set as
	select r.CusipObservationMonth, r.c_0, r.c as factor
	from regression as r
	where r._TYPE_ = 'FINAL';
	quit;

	/*get average factor and scale final data sets such that the average factor equals 1 (see C.1)*/
	%global avg_fac;
	proc sql noprint;
	select distinct avg(factor)
	into :avg_fac
	from Final_set;
	quit;
	%put &avg_fac;

	proc sql;
	create table Final_set as
	select CusipObservationMonth, c_0, factor/&avg_fac as factor
	from Final_set;
	quit;

	proc sql;
	create table regression_1st_res as
	select entrd_vol_qt, c_vol*&avg_fac as c_vol
	from regression_1st_res;
	quit;

	proc sql;
	create table sum_squares as
	select sum(_SSE_)  as sum_sse, count(*) as number_of_observations
	from regression y
	where y._type_ = 'FINAL';
	quit;

	data convergence_Schultz_&cur_year._&cur_mth.;
	set sum_squares;
	run;

	/*call the iteration macro*/
	%start_iteration();

	/*add functional from to data set*/
	proc sql;
	create table sizeAdap.Schultz_c_vol_iter as
	select a.*, b.c_vol as c_vol_&cur_year._&cur_mth.
	from sizeAdap.Schultz_c_vol_iter a, regression_1st_res b
	where a.entrd_vol_qt=b.entrd_vol_qt;
	quit;

	/*add factors to data set*/
	proc append base=sizeAdap.Schultz_monthly data=Final_set;quit;

	/*update year and mth*/
	%if &cur_mth=12 %then %do;
		%let cur_year=%eval(&cur_year+1);
		%let cur_mth=1;
	%end;
	%else %do;
		%let cur_mth=%eval(&cur_mth+1);
	%end;

%end;

%mend monthly;

*call macro;
%monthly(2004,10);

%global note;
%let note='Iteration finished';
%put &note;


/*Calculate normalization factor*/
proc sql;
create table vol_weight as
select distinct year(trd_exctn_dt) as year, month(trd_exctn_dt) as month, min(entrd_vol_qt,20000000) as entrd_vol_qt,
	sum(min(entrd_vol_qt,20000000)) as sum_vol
from sizeAdap.Schultz_Sample
group by calculated year, calculated month, calculated entrd_vol_qt;
quit;
proc sql;
create table vol_weight_2 as
select *, sum_vol/sum(sum_vol) as weight_vol
from vol_weight
group by year, month;
quit;

proc transpose data=sizeAdap.schultz_c_vol_iter out=c_vol_long (rename=(col1=c_vol)); by entrd_vol_qt;quit;

proc sql;
create table sizeAdap.Schultz_norm_fact_iter as
select distinct v.year, v.month, sum(v.weight_vol*f.c_vol) as normalization_factor
from vol_weight_2 v, c_vol_long f
where v.year=input(substr(f._NAME_,7,4),4.) and v.month=input(substr(f._NAME_,12,2),2.) and v.entrd_vol_qt=f.entrd_vol_qt
group by v.year, v.month;
quit;

*Save final results; 
proc sql;
create table sizeAdap.Schultz_b_size_results as
select r.CusipObservationMonth, r.factor * n.normalization_factor as SizeAdap_Schultz
from sizeAdap.Schultz_monthly as r, SizeAdap.Schultz_norm_fact_iter n
where input(substr(r.cusipObservationMonth,11,4),4.)=n.year and  input(substr(r.cusipObservationMonth,16,2),2.)=n.month
order by CusipObservationMonth;
quit; 
